package gov.usgs.earthquake.nshmp.data;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.stream.DoubleStream;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import com.google.common.primitives.Doubles;

/**
 * An immutable sequence of xy-value pairs that is iterable ascending in x.
 * 
 * <p>For mutable y-values, use the {@link MutableXySequence} sub-type. Methods
 * and classes in this package will always return a {@link MutableXySequence} if
 * the instance is, in fact, mutable. Mutable sequences should not be considered
 * thread safe.
 * 
 * <p>All data supplied to the factory methods in this interface is copied
 * unless it is not necessary to do so. Users should consider
 * {@link #copyOf(XySequence)} where possible as x-values will never be
 * replicated in memory.
 * 
 * <p>Streams and iterators over an XySequence return points for which
 * {@link XyPoint#set(double)} throws an {@code UnsupportedOperationException}
 * as does {@code Iterator.remove()}.
 *
 * @author Peter Powers
 * @see MutableXySequence
 */
public interface XySequence extends Iterable<XyPoint> {

  // TODO better spliterator?
  // TODO add index to XyPoint?

  /**
   * Create a new, immutable sequence from the supplied value arrays.
   *
   * @param xs x-values to initialize sequence with
   * @param ys y-values to initialize sequence with
   * @throws IllegalArgumentException if {@code xs} is empty
   * @throws IllegalArgumentException if {@code xs} and {@code ys} are not the
   *         same size
   * @throws IllegalArgumentException if {@code xs} does not increase
   *         monotonically or contains repeated values
   */
  static XySequence create(
      double[] xs,
      double[] ys) {

    return Sequences.construct(
        Arrays.copyOf(xs, xs.length),
        Arrays.copyOf(ys, ys.length));
  }

  /**
   * Create a new, immutable sequence from the supplied value collections.
   *
   * @param xs x-values to initialize sequence with
   * @param ys y-values to initialize sequence with
   * @throws IllegalArgumentException if {@code xs} is empty
   * @throws IllegalArgumentException if {@code xs} and {@code ys} are not the
   *         same size
   * @throws IllegalArgumentException if {@code xs} does not increase
   *         monotonically or contains repeated values
   */
  static XySequence create(
      Collection<? extends Number> xs,
      Collection<? extends Number> ys) {

    return Sequences.construct(
        Doubles.toArray(xs),
        Doubles.toArray(ys));
  }

  /**
   * Create an immutable copy of the supplied sequence.
   *
   * @param sequence to copy
   */
  static XySequence copyOf(XySequence sequence) {
    /* if immutable, return argument */
    return (sequence.getClass().equals(ArrayXySequence.class))
        ? sequence
        : new ArrayXySequence(sequence, false);
  }

  /**
   * Adds a {@code sequence} to the exisiting sequence for {@code key} in the
   * supplied {@code map}. If the key is not present in the map, this method
   * puts a mutable copy of the supplied sequence in the map.
   * 
   * @param key for sequence to add
   * @param map of sequences to add to
   * @throws IllegalArgumentException if the x-values of added sequences do not
   *         match those of existing sequences
   */
  static <E extends Enum<E>> void addToMap(
      E key,
      Map<E, MutableXySequence> map,
      XySequence sequence) {

    if (map.containsKey(key)) {
      map.get(key).add(sequence);
    } else {
      map.put(key, MutableXySequence.copyOf(sequence));
    }
  }

  /**
   * Returns the x-value at index.
   * 
   * @param index to retrieve
   * @throws IndexOutOfBoundsException if the index is out of range
   */
  double x(int index);

  /**
   * Returns the y-value at index.
   * 
   * @param index to retrieve
   * @throws IndexOutOfBoundsException if the index is out of range
   */
  double y(int index);

  /**
   * Returns a double stream of the sequence x-values.
   */
  DoubleStream xValues();

  /**
   * Returns a double stream of the sequence y-values.
   */
  DoubleStream yValues();

  /**
   * Returns a stream of the {@link XyPoint}s in this sequence.
   */
  default Stream<XyPoint> stream() {
    return StreamSupport.stream(spliterator(), false);
  }

  /**
   * Returns the number of points in this sequence.
   */
  int size();

  /**
   * Returns the first {@link XyPoint} in this sequence.
   */
  XyPoint min();

  /**
   * Returns the last {@link XyPoint} in this sequence.
   */
  XyPoint max();

  /**
   * Returns {@code true} if all y-values are 0.0; {@code false} otherwise.
   */
  boolean isClear();

  /**
   * Returns a new sequence that has had all leading and trailing points for
   * which {@code y = 0} removed. Any {@code y = 0} valued points in the middle
   * of this sequence are ignored.
   * 
   * @throws IllegalStateException if {@link #isClear() this.isClear()} as empty
   *         sequences are not permitted
   */
  XySequence trim();

}
