package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.Preconditions.checkState;

import java.util.Optional;
import java.util.OptionalDouble;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Magnitude frequency distribution (MFD) configuration. Epistemic uncertainty
 * and aleatory variability on maximum magnitude. If any epistemic {@code mMax <
 * minMagnitude} uncertainty is not applied.
 * 
 * @author U.S. Geological Survey
 */
class MfdConfig {

  static final String EPISTEMIC_TREE_KEY = "epistemic-tree";
  static final String ALEATORY_PROPS_KEY = "aleatory-properties";
  static final String MIN_MAGNITUDE_KEY = "minimum-magnitude";

  final Optional<LogicTree<Double>> epistemicTree;
  final Optional<AleatoryProperties> aleatoryProperties;
  final OptionalDouble minMagnitude;

  private MfdConfig(Builder builder) {
    this.epistemicTree = builder.epistemicTree;
    this.aleatoryProperties = builder.aleatoryProperties;
    this.minMagnitude = builder.minMagnitude;
  }

  static MfdConfig empty() {
    return builder().build();
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static class Builder {

    private Optional<LogicTree<Double>> epistemicTree = Optional.empty();
    private Optional<AleatoryProperties> aleatoryProperties = Optional.empty();
    private OptionalDouble minMagnitude = OptionalDouble.empty();
    private boolean built = false;

    Builder epistemicTree(LogicTree<Double> tree) {
      this.epistemicTree = Optional.ofNullable(tree);
      return this;
    }

    Builder aleatoryProperties(AleatoryProperties props) {
      this.aleatoryProperties = Optional.ofNullable(props);
      return this;
    }

    Builder minimumMagnitude(double mMin) {
      Earthquakes.checkMagnitude(mMin);
      this.minMagnitude = OptionalDouble.of(mMin);
      return this;
    }

    MfdConfig build() {
      checkState(!built, "This MfdConfig.Builder has already been used");
      if (epistemicTree.isPresent() || aleatoryProperties.isPresent()) {
        checkState(minMagnitude.isPresent(), "%s required", EPISTEMIC_TREE_KEY);
      }
      built = true;
      return new MfdConfig(this);
    }
  }

  static class AleatoryProperties {

    final int count;
    final boolean momentBalanced;
    final int σSize;
    final double σ;

    AleatoryProperties(
        int count,
        boolean momentBalanced,
        int σSize,
        double σ) {

      this.count = count;
      this.momentBalanced = momentBalanced;
      this.σSize = σSize;
      this.σ = σ;
    }
  }
}
