package gov.usgs.earthquake.nshmp.mfd;

import static gov.usgs.earthquake.nshmp.Earthquakes.checkMagnitude;
import static gov.usgs.earthquake.nshmp.Earthquakes.magToMoment;
import static gov.usgs.earthquake.nshmp.mfd.Mfds2.gutenbergRichterRate;

import java.util.Arrays;
import java.util.Optional;
import java.util.function.Consumer;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.data.XySequence;

/**
 * Magnitude frequency distribution (MFD) builder. A builder may be obtained via
 * the factory methods in {@link Mfds}. {@code MfdBuilder} is an intermediate
 * class that can modify and shape an MFD prior to calling {@link #build()},
 * which returns an immutable {@link Mfd}.
 * 
 * @author Peter Powers
 * @see Mfds
 */
public class MfdBuilder {

  private final MutableXySequence mfd;

  /*
   * TODO Developer notes:
   * 
   * newSingleMfdTree(json) // e.g. RLMEs explicit magnitudes, rates, wts
   * 
   * should be able to leverage NOT copying x-values/magnitudes via copy
   * constructor
   * 
   * applyEpistemicUncertainty?? return treeBuilder?? may have to have
   * individuak builders that store parameters
   * 
   * check: is min=max allowed; check docs as well
   * 
   * add constraints preconditions for params other than magnitude
   */

  MfdBuilder(double[] magnitudes, Optional<double[]> rates) {
    this.mfd = MutableXySequence.create(magnitudes, rates);
  }

  /* Single MFD */
  static MfdBuilder initSingle(
      double magnitude,
      double rate) {

    checkMagnitude(magnitude);
    return new MfdBuilder(
        new double[] { magnitude },
        Optional.of(new double[] { rate }));
  }

  /* Incremental MFD */
  static MfdBuilder initIncremental(
      double[] magnitudes,
      double[] rates) {

    Arrays.stream(magnitudes).forEach(Earthquakes::checkMagnitude);
    return new MfdBuilder(magnitudes, Optional.of(rates));
  }

  /* Gaussian MFD */
  static MfdBuilder initGaussian(
      double μ,
      double σ,
      double n) {

    checkMagnitude(μ);
    double mMin = checkMagnitude(μ - 2 * σ);
    double mMax = checkMagnitude(μ - 2 * σ);
    double Δm = (mMax - mMin) / n;
    double[] magnitudes = DoubleData.buildCleanSequence(mMin, mMax, Δm, true, 5);
    MfdBuilder builder = new MfdBuilder(magnitudes, Optional.empty());
    builder.mfd.forEach(p -> p.set(Maths.normalPdf(μ, σ, p.x())));
    return builder;
  }

  /* Gutenberg–Richter MFD */
  static MfdBuilder initGutenbergRichter(
      double mMin,
      double mMax,
      double Δm,
      double b) {

    double[] magnitudes = DoubleData.buildCleanSequence(
        checkMagnitude(mMin),
        checkMagnitude(mMax),
        Δm, true, 3);
    MfdBuilder builder = new MfdBuilder(magnitudes, Optional.empty());
    builder.mfd.forEach(p -> p.set(gutenbergRichterRate(0, b, p.x())));
    return builder;
  }

  /* GUtneberg–Richter with exponential taper */
  static MfdBuilder initTaperedGutenbergRichter(
      double mMin,
      double mMax,
      double Δm,
      double b,
      double mc) {

    double[] magnitudes = DoubleData.buildCleanSequence(
        checkMagnitude(mMin),
        checkMagnitude(mMax),
        Δm, true, 3);
    MfdBuilder builder = new MfdBuilder(magnitudes, Optional.empty());
    GrTaper grTaper = new GrTaper(Δm, b, mc);
    builder.mfd.forEach(p -> p.set(
        gutenbergRichterRate(0, b, p.x()) *
            grTaper.scale(p.x())));
    return builder;
  }

  private static final double M_MIN_MOMENT = magToMoment(4.0);
  private static final double M_MAX_MOMENT = magToMoment(9.05);

  /* Gutenberg–Richter tapering scale function */
  private static class GrTaper {

    final double ΔmBy2;
    final double β;
    final double Mcm;

    GrTaper(double Δm, double b, double mc) {
      ΔmBy2 = Δm / 2.0;
      β = b / 1.5;
      Mcm = magToMoment(mc);
    }

    /* Resolve bin edges and return scale factor */
    double scale(double m) {
      double Mlo = magToMoment(m - ΔmBy2);
      double Mhi = magToMoment(m + ΔmBy2);
      double binRateTapered = paretoΔ(M_MIN_MOMENT, Mlo, Mhi, β, Mcm);
      double binRateUntapered = paretoΔ(M_MIN_MOMENT, Mlo, Mhi, β, M_MAX_MOMENT);
      return binRateTapered / binRateUntapered;
    }
  }

  /*
   * Computes the number of events in a tapered GR magnitude bin. All 'M' values
   * are scalar seismic moment.
   */
  private static double paretoΔ(
      double Mt,
      double Mlo,
      double Mhi,
      double β,
      double Mcm) {

    return Mfds2.paretoRate(Mt, Mlo, β, Mcm) - Mfds2.paretoRate(Mt, Mhi, β, Mcm);
  }

  /**
   * Returns an newly-created immutable {@code Mfd}.
   */
  public XySequence build() {
    return XySequence.copyOf(mfd);
  }

  /**
   * Scales the MFD being built to the supplied cumulative rate.
   * 
   * @param cumulativeRate to scale MFD to
   * @return {@code this} builder, for use inline
   */
  public MfdBuilder scaleToCumulativeRate(double cumulativeRate) {
    return this;
  }

  /**
   * Scales the MFD being built to the supplied total moment rate.
   * 
   * @param momentRate to scale MFD to
   * @return {@code this} builder, for use inline
   */
  public MfdBuilder scaleToMomentRate(double momentRate) {
    return this;
  }

  /**
   * Transforms the rates in the MFD being built by applying the supplied action
   * to each {@link XyPoint}.
   * 
   * @param action to perform on each MFD element
   * @return {@code this} builder, for use inline
   */
  public MfdBuilder transform(Consumer<XyPoint> action) {
    mfd.transform(action);
    return this;
  }

}
