package gov.usgs.earthquake.nshmp.mfd;

/**
 * Utility methods for creating and working with magnitude frequency
 * distributions (MFDs). The factory methods of this class all return an
 * {@link MfdBuilder} that may be used to modify and shape an MFD prior to
 * calling {@link MfdBuilder#build()}, which returns an immutable {@link Mfd}.
 * For example:
 * 
 * <pre>
 * Mfd mfd = Mfds.newGutenbergRighterMfd(5.05, 7.25, 0.1, 1.0)
 *     .scaleToCumulativeRate(0.02)
 *     .transform(p -> p.set((p.x() > 6.5) ? p.y() * 0.667 : p.y()))
 *     .build()
 * </pre>
 * 
 * In the example above, an incremental MFD with a <a
 * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law"
 * target="_top">Gutenberg–Richter</a> distribution of rates is initialized from
 * {@code M=5.05} to {@code M=7.25} with a step size of {@code 0.1} magnitude
 * units. The resultant magnitude array,
 * {@code [5.05, 5.15, 5.25, ... , 7.15, 7.25]}, is typically used to define the
 * centers of 0.1 magnitude intervals spanning the magnitude range
 * {@code [5.0..7.3]}. The MFD is then scaled to the cumulative rate of 0.02
 * events/year, and finally those rates above {@code M=6.5} are scaled to ⅔ of
 * their present value. Note that an {@code MfdBuilder} executes methods in the
 * order defined, so reversing {@code scaleToCumulativeRate()} and
 * {@code transform()} in the example above results in two different MFDs.
 * 
 * @author Peter Powers
 * @see MfdBuilder
 * @see Mfd
 */
public interface Mfds2 {

  /**
   * Create an {@code MfdBuilder} for a single magnitude.
   * 
   * @param magnitude
   * @param rate
   * @throws IllegalArgumentException if magnitude is outside the range
   *         {@code [-2.0..9.7]}
   */
  public static MfdBuilder newSingleMfd(double magnitude, double rate) {
    return MfdBuilder.initSingle(magnitude, rate);
  }

  /**
   * Create an {@code MfdBuilder} for multiple magnitudes with explicitly
   * defined rates.
   * 
   * @param magnitudes
   * @param rates
   * @throws IllegalArgumentException if {@code magnitudes} is empty
   * @throws IllegalArgumentException if {@code magnitudes} and {@code rates}
   *         are not the same size
   * @throws IllegalArgumentException if {@code magnitudes} does not increase
   *         monotonically or contains repeated values
   * @throws IllegalArgumentException if any magnitude is outside the range
   *         {@code [-2.0..9.7]}
   */
  public static MfdBuilder newIncrementalMfd(
      double[] magnitudes,
      double[] rates) {

    return MfdBuilder.initIncremental(magnitudes, rates);
  }

  /**
   * Create an {@code MfdBuilder} for a doubly-truncated Gaussian MFD. Gaussian
   * MFDs are typically used to represent aleatory variability in magnitude.
   * This implementation adheres to the NSHMP standard of using a normal
   * distribution truncated at {@code ±2σ}. When building the underlying
   * magnitude distribtion, this implementation rounds each magnitude to the 5
   * decimal places.
   * 
   * @param μ the mean magnitude
   * @param σ the standard deviation
   * @param n the number of points in the distribution
   * @throws IllegalArgumentException if {@code μ} is outside the range
   *         {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code σ} is outside the range
   *         {@code (0..0.5]}
   */
  public static MfdBuilder newGaussianMfd(
      double μ,
      double σ,
      double n) {

    return MfdBuilder.initGaussian(μ, σ, n);
  }

  /**
   * Create an {@code MfdBuilder} for a truncated <a
   * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law"
   * target="_top">Gutenberg–Richter</a> MFD. When building the underlying
   * magnitude distribtion, this implementation rounds each magnitude to the 3
   * decimal places.
   * 
   * @param mMin the minimum truncation magnitude
   * @param mMax the maximum truncation magnitude
   * @param Δm the magnitude step of the distribtion
   * @param b the Gutenberg-Richter b-value
   * @throws IllegalArgumentException if {@code mMin} or {@code mMax} is outside
   *         the range {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code mMin > mMax}
   * @throws IllegalArgumentException if {@code Δm > mMax - mMin}
   * @throws IllegalArgumentException if {@code b} is outside the range
   *         {@code [0..2]}
   */
  public static MfdBuilder newGutenbergRichterMfd(
      double mMin,
      double mMax,
      double Δm,
      double b) {

    return MfdBuilder.initGutenbergRichter(mMin, mMax, Δm, b);
  }

  /**
   * Create an {@code MfdBuilder} for a truncated <a
   * href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law"
   * target="_top">Gutenberg–Richter</a> MFD with an exponential taper. This
   * implementation uses the scalar seismic moment based distribution of Kagan
   * (2002), Equation 11 (see
   * {@link #paretoRate(double, double, double, double)} for reference) to
   * compute scale factors for each magnitude bin that are then applied to a
   * standard Gutenberg–Richter distribution. When building the underlying
   * magnitude distribtion, this implementation rounds each magnitude to the 3
   * decimal places.
   * 
   * @param mMin the minimum truncation magnitude
   * @param mMax the maximum truncation magnitude
   * @param Δm the magnitude step of the distribtion
   * @param b the Gutenberg-Richter b-value
   * @param mc the corner magnitude of the distribution
   * @throws IllegalArgumentException if {@code mMin}, {@code mMax}, or
   *         {@code mc} is outside the range {@code [-2.0..9.7]}
   * @throws IllegalArgumentException if {@code mMin > mMax} or {@code mc} is
   *         not {@code mMin < mc < mMax}
   * @throws IllegalArgumentException if {@code Δm > mMax – mMin}
   * @throws IllegalArgumentException if {@code b} is outside the range
   *         {@code [0..2]}
   */
  public static MfdBuilder newTaperedGutenbergRichterMfd(
      double mMin,
      double mMax,
      double Δm,
      double b,
      double mc) {

    return MfdBuilder.initTaperedGutenbergRichter(mMin, mMax, Δm, b, mc);
  }

  /**
   * Returns the <a href="https://en.wikipedia.org/wiki/Gutenberg–Richter_law"
   * target="_top">Gutenberg–Richter</a> event rate for the supplied a- and
   * b-values and magnitude. This method follows the form
   * {@code N(M) = 10<sup>(a – bM)</sup>}. If one's {@code a}-values are
   * represented as the total number of events {@code > M}, sometimes denoted as
   * {@code A} and as NSHM a-grid values commonly are, be sure to supply
   * {@code log10(A)} instead.
   *
   * <p><b>Reference:</b> Gutenberg, B., Richter, C. F., 1956, Magnitude and
   * Energy of Earthquakes, Annali di Geofisica, v. 9, p. 1–15, doi:<a
   * href="https://doi.org/10.4401/ag-5590" target=_top">10.4401/ag-5590<a>
   * 
   * @param a value (log10 rate of M=0 events)
   * @param b value
   * @param M magnitude of interest
   * @return the rate of magnitude {@code M} events
   */
  public static double gutenbergRichterRate(double a, double b, double M) {
    return Math.pow(10, a - b * M);
  }

  /**
   * Returns the cumulative number of events with seismic moment greater than
   * {@code M} with an exponential taper applied. This is the complementary
   * Pareto distribution as recorded in Equation 11 of Kagan (2002).
   * 
   * <p><b>Reference:</b> Kagan, Y.Y., 2002, Seismic moment distribution
   * revisited: I. Statistical results, Geophysical Journal International, v.
   * 148, p. 520—541, doi:<a
   * href="https://doi.org/10.1046/j.1365-246x.2002.01594.x"
   * target=_top">10.1046/j.1365-246x.2002.01594.x<a>
   * 
   * @param Mt the scalar seismic moment of catalog completeness threshold
   * @param M the scalar seismic moment of the target magnitude
   * @param β the index parameter of the distribution; {@code β = ⅔b}
   *        (Gutenberg–Richter {@code b}-value)
   * @param Mcm the scalar seismic moment of the upper bound corner magnitude
   */
  public static double paretoRate(
      double Mt,
      double M,
      double β,
      double Mcm) {

    return Math.pow(Mt / M, β) * Math.exp((Mt - M) / Mcm);
  }
}
