package gov.usgs.earthquake.nshmp.tree;

import java.util.List;

/**
 * A logic tree.
 * 
 * <p>Factory methods are provided that return builders for specialized branch
 * implementations. Note that the precision of weight values supplied to
 * {@code Builder.addBranch()} methods should be limited to 6 decimal places.
 */
public interface LogicTree<T> extends Iterable<Branch<T>> {

  /**
   * Return an immutable list of logic tree branches.
   */
  List<Branch<T>> branches();

  /**
   * Return a logic tree branch corresponding to the supplied probability.
   * 
   * <p><b>Note:</b> this method does not check that {@code 0.0 ≤ P < 1.0}. For
   * {@code P < 0.0}, method returns the first branch of the tree; for
   * {@code P >= 1.0} method returns the last branch of the tree.
   * 
   * @param probability in the range [0..1)
   */
  Branch<T> sample(double probability);

  /**
   * Return an immutable list of logic tree branches corresponding to the
   * supplied probabilities.
   * 
   * <p><b>Note:</b> this method does not check that each {@code 0.0 ≤ P < 1.0}.
   * For {@code P < 0.0}, method returns the first branch of the tree; for
   * {@code P >= 1.0} method returns the last branch of the tree.
   * 
   * @param probabilities array with values in the range [0..1)
   */
  List<Branch<T>> sample(double[] probabilities);

  /**
   * Return a new, single-use {@code LogicTree} builder.
   */
  static <T> Builder<T> builder() {
    return new Builder<T>();
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} of the built tree is an {@link Enum#name()} and each
   * {@link Branch#value()} is any value type.
   */
  static <E extends Enum<E>, T> EnumBuilder<E, T> enumBuilder() {
    return new EnumBuilder<E, T>();
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} of the built tree is an {@link Enum#name()} and each
   * {@link Branch#value()} is the enum value itself.
   */
  static <E extends Enum<E>> EnumValueBuilder<E> enumValueBuilder() {
    return new EnumValueBuilder<E>();
  }

  /**
   * Return a new, single-use {@code LogicTree} builder where each
   * {@link Branch#id()} and {@link Branch#value()} of the built tree is the
   * same string.
   */
  static StringValueBuilder stringValueBuilder() {
    return new StringValueBuilder();
  }

  /**
   * A single-use {@code LogicTree} builder.
   */
  static class Builder<T> extends TreeBuilder<T> {

    private Builder() {}

    /**
     * Add a branch to the tree.
     * 
     * @param id of the branch
     * @param value of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} or {@code value} is null
     * @throws IllegalArgumentException if {@code id} is empty or whitespace
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public Builder<T> addBranch(String id, T value, double weight) {
      branches.add(new RegularBranch<>(id, value, weight));
      return this;
    }

    /**
     * Return a newly created {@code LogicTree}.
     * 
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     * @throws IllegalStateException if branch weights do not sum to
     *         {@code 1.0 ± 0.00001}
     */
    @Override
    public LogicTree<T> build() {
      return super.build();
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} of
   * the built tree is an {@link Enum#name()} and each {@link Branch#value()} is
   * any value type.
   */
  static class EnumBuilder<E extends Enum<E>, T> extends TreeBuilder<T> {

    private EnumBuilder() {}

    /**
     * Add a branch to the tree.
     * 
     * @param id of the branch
     * @param value of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} or {@code value} is null
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public EnumBuilder<E, T> addBranch(E id, T value, double weight) {
      branches.add(new EnumBranch<E, T>(id, value, weight));
      return this;
    }

    /**
     * Return a newly created {@code LogicTree}.
     * 
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     * @throws IllegalStateException if branch weights do not sum to
     *         {@code 1.0 ± 0.00001}
     */
    @Override
    public LogicTree<T> build() {
      return super.build();
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} of
   * the built tree is an {@link Enum#name()} and each {@link Branch#value()} is
   * the enum value itself.
   */
  static class EnumValueBuilder<E extends Enum<E>> extends TreeBuilder<E> {

    private EnumValueBuilder() {}

    /**
     * Add a branch to the tree.
     * 
     * @param id of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} is null
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public EnumValueBuilder<E> addBranch(E id, double weight) {
      branches.add(new EnumValueBranch<E>(id, weight));
      return this;
    }

    /**
     * Return a newly created {@code LogicTree}.
     * 
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     * @throws IllegalStateException if branch weights do not sum to
     *         {@code 1.0 ± 0.00001}
     */
    @Override
    public LogicTree<E> build() {
      return super.build();
    }
  }

  /**
   * A single-use {@code LogicTree} builder where each {@link Branch#id()} and
   * {@link Branch#value()} of the built tree is the same string.
   */
  static class StringValueBuilder extends TreeBuilder<String> {

    private StringValueBuilder() {}

    /**
     * Add a branch to the tree.
     * 
     * @param id of the branch
     * @param weight of the branch
     * @return this {@code Builder} object
     * @throws NullPointerException if {@code id} is null
     * @throws IllegalArgumentException if {@code id} is empty or whitespace
     * @throws IllegalArgumentException if {@code weight} is outside the range
     *         (0..1]
     */
    public StringValueBuilder addBranch(String id, double weight) {
      branches.add(new StringValueBranch(id, weight));
      return this;
    }

    /**
     * Return a newly created {@code LogicTree}.
     * 
     * @throws IllegalStateException if this method has already been called
     * @throws IllegalStateException if the logic tree is empty (i.e. no
     *         branches have been added)
     * @throws IllegalStateException if branch weights do not sum to
     *         {@code 1.0 ± 0.00001}
     */
    @Override
    public LogicTree<String> build() {
      return super.build();
    }
  }
}
