package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Text;

@SuppressWarnings("javadoc")
public class XySequenceTests {

  /* Tests XySequence and array backed implementation. */

  private static double[] xs = new double[] { 0, 1, 2, 3 };
  private static double[] ys = new double[] { -1, 10.5, 5.25, 2.5 };
  private static XySequence xy = XySequence.create(xs, ys);

  @Test
  public final void factoryConstructorTests() {
    // arrays
    assertArrayEquals(xs, ((ArrayXySequence) xy).xs, 0);
    assertArrayEquals(ys, ((ArrayXySequence) xy).ys, 0);
    // collections; guava Doubles copies array on toArray()
    List<Double> xList = Doubles.asList(xs);
    List<Double> yList = Doubles.asList(ys);
    XySequence xy2 = XySequence.create(xList, yList);
    assertArrayEquals(xs, ((ArrayXySequence) xy2).xs, 0);
    assertArrayEquals(ys, ((ArrayXySequence) xy2).ys, 0);

    XySequence xy3 = XySequence.copyOf(xy);
    assertSame(xy, xy3);

    // tests that a mutable sequence passed to the XySequence
    // constructor is copied
    MutableXySequence xy4 = MutableXySequence.copyOf(xy);
    XySequence xy5 = XySequence.copyOf(xy4);
    assertNotSame(xy4, xy5);

    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.create(new double[0], new double[0]);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      XySequence.create(xs, new double[0]);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[] reversed = Arrays.copyOf(xs, xs.length);
      Doubles.reverse(reversed);
      XySequence.create(reversed, ys);
    });
  }

  private enum MapKey {
    K1;
  }

  @Test
  public final void addToMapTest() {

    Map<MapKey, MutableXySequence> testMap = new HashMap<>();

    // test adding a mutable copy to map for non-existant key
    XySequence.addToMap(MapKey.K1, testMap, xy);
    assertNotSame(xy, testMap.get(MapKey.K1));
    assertTrue(testMap.get(MapKey.K1) instanceof MutableArrayXySequence);

    // test adding to existing key
    XySequence.addToMap(MapKey.K1, testMap, xy);
    double[] sum = DoubleData.add(Arrays.copyOf(ys, ys.length), ys);
    assertArrayEquals(sum, ((MutableArrayXySequence) testMap.get(MapKey.K1)).ys, 0);

    assertThrows(IllegalArgumentException.class, () -> {
      Map<MapKey, MutableXySequence> map = new HashMap<>();
      XySequence.addToMap(MapKey.K1, map, xy);
      XySequence xyNonDuplicateX = XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
      XySequence.addToMap(MapKey.K1, map, xyNonDuplicateX);
    });
  }

  @Test
  public final void pointSetTest() {
    assertThrows(UnsupportedOperationException.class, () -> {
      xy.stream().forEach(p -> p.set(0));
    });
  }

  @Test
  public final void xyTest() {
    for (int index = 0; index < xy.size(); index++) {
      assertEquals(xs[index], xy.x(index), 0);
      assertEquals(ys[index], xy.y(index), 0);
    }

    assertThrows(IndexOutOfBoundsException.class, () -> {
      xy.x(10);
    });

    assertThrows(IndexOutOfBoundsException.class, () -> {
      xy.y(10);
    });
  }

  @Test
  public final void xyValuesTest() {
    assertArrayEquals(xs, xy.xValues().toArray(), 0);
    assertArrayEquals(ys, xy.yValues().toArray(), 0);
  }

  @Test
  public final void streamTest() {
    assertArrayEquals(xs, xy.stream().mapToDouble(XyPoint::x).toArray(), 0);
    assertArrayEquals(ys, xy.stream().mapToDouble(XyPoint::y).toArray(), 0);
  }

  @Test
  public final void sizeTest() {
    assertEquals(xs.length, xy.size());
    assertEquals(ys.length, xy.size());
  }

  @Test
  public final void minMaxTest() {
    XyPoint first = xy.min();
    assertEquals(xs[0], first.x(), 0);
    assertEquals(ys[0], first.y(), 0);
    XyPoint last = xy.max();
    assertEquals(xs[xs.length - 1], last.x(), 0);
    assertEquals(ys[ys.length - 1], last.y(), 0);
    // test correct class (immutable)
    assertEquals(first.getClass(), ArrayXySequence.Point.class);
    assertEquals(last.getClass(), ArrayXySequence.Point.class);
  }

  @Test
  public final void isClearTest() {
    assertFalse(xy.isClear());
    XySequence xy = XySequence.create(
        new double[] { 1, 2, 3 },
        new double[3]);
    assertTrue(xy.isClear());
  }

  @Test
  public final void trimTest() {
    double[] xsOriginal = { 0, 1, 2, 3, 4, 5 };
    double[] ysOriginal = { 0, 0, 1, 2, 0, 0 };
    double[] xsExpected = { 2, 3 };
    double[] ysExpected = { 1, 2 };
    XySequence xyExpected = XySequence.create(xsExpected, ysExpected);
    XySequence xyTrimmed = XySequence.create(xsOriginal, ysOriginal).trim();
    assertEquals(xyExpected, xyTrimmed);
    // test class type
    assertEquals(xyTrimmed.getClass(), ArrayXySequence.class);

    assertThrows(IllegalStateException.class, () -> {
      XySequence xy = XySequence.create(
          new double[] { 1, 2, 3 },
          new double[3]);
      xy.trim();
    });
  }

  @Test
  public final void testEquals() {
    XySequence xyCopy = XySequence.copyOf(xy);
    // test ==
    assertTrue(xyCopy.equals(xy));
    // test instanceof
    assertFalse(xy.equals(new Object()));
    // test identical data
    XySequence xyDuplicate = XySequence.create(xs, ys);
    assertTrue(xy.equals(xyDuplicate));
    // test non-identical data
    XySequence xyNonDuplicateX = XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
    assertFalse(xy.equals(xyNonDuplicateX));
    XySequence xyNonDuplicateY = XySequence.create(xs, new double[xs.length]);
    assertFalse(xy.equals(xyNonDuplicateY));
  }

  @Test
  public final void testHashCode() {
    int hashActual = xy.hashCode();
    int hashExpected = Arrays.deepHashCode(new Object[] { xs, ys });
    assertEquals(hashExpected, hashActual);
  }

  @Test
  public final void testValidateSequence() {
    // testing hash mismatch but matching values succeeds
    ArrayXySequence xyAltHashAndValues =
        (ArrayXySequence) XySequence.create(new double[] { 0, 1, 2, 3 }, ys);
    ArrayXySequence axy = (ArrayXySequence) xy;
    axy.validateSequence(xyAltHashAndValues);

    assertThrows(IllegalArgumentException.class, () -> {
      // casting required for internal methods
      ArrayXySequence xyAltHash =
          (ArrayXySequence) XySequence.create(new double[] { 1, 2, 3, 4 }, ys);
      ArrayXySequence axy2 = (ArrayXySequence) xy;
      axy2.validateSequence(xyAltHash);
    });
  }

  @Test
  public final void testToString() {
    String expected = new StringBuilder()
        .append("ArrayXySequence:").append(Text.NEWLINE)
        .append("XyPoint: [0.0, -1.0]").append(Text.NEWLINE)
        .append("XyPoint: [1.0, 10.5]").append(Text.NEWLINE)
        .append("XyPoint: [2.0, 5.25]").append(Text.NEWLINE)
        .append("XyPoint: [3.0, 2.5]").toString();
    assertEquals(expected, xy.toString());
  }

  @Test
  public final void testIteratorRemoveUOE() {
    assertThrows(UnsupportedOperationException.class, () -> {
      xy.iterator().remove();
    });
  }

}
