package gov.usgs.earthquake.nshmp.geo.json;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.List;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;

public class GeometryTests {

  private static final Location POINT_LOC = Location.create(34, -117);

  private static final LocationList LINE_LOCS = LocationList.of(
      Location.create(34, -117),
      Location.create(35, -118),
      Location.create(37, -116),
      Location.create(38, -117));

  private static final List<LocationList> MULTI_LINE_LOCS = Lists.newArrayList(
      LINE_LOCS,
      LocationList.of(
          Location.create(44, -117),
          Location.create(45, -118),
          Location.create(47, -116),
          Location.create(48, -117)));

  private static final double[] POINT_COORDS =
      new double[] { POINT_LOC.longitude, POINT_LOC.latitude };

  private static final double[][] LINE_COORDS = new double[LINE_LOCS.size()][];

  private static final double[][][] MULTI_LINE_COORDS =
      new double[MULTI_LINE_LOCS.size()][LINE_LOCS.size()][];

  private static final Geometry<double[]> POINT = new Geometry<>(Type.POINT, POINT_COORDS);

  private static final Geometry<double[][]> LINE_STRING =
      new Geometry<>(Type.LINE_STRING, LINE_COORDS);

  private static final Geometry<double[][][]> MULTI_LINE_STRING =
      new Geometry<>(Type.MULTI_LINE_STRING, MULTI_LINE_COORDS);

  private static final Geometry<double[][][]> POLYGON =
      new Geometry<>(Type.POLYGON,
          MULTI_LINE_COORDS);

  static {
    for (int i = 0; i < LINE_LOCS.size(); i++) {
      LINE_COORDS[i] = new double[] {
          LINE_LOCS.get(i).longitude,
          LINE_LOCS.get(i).latitude };
    }

    for (int i = 0; i < MULTI_LINE_LOCS.size(); i++) {
      LocationList locs = MULTI_LINE_LOCS.get(i);
      for (int j = 0; j < locs.size(); j++) {
        MULTI_LINE_COORDS[i][j] = new double[] {
            locs.get(j).longitude,
            locs.get(j).latitude
        };
      }
    }
  }

  @Nested
  @DisplayName("Point Tests")
  public static class PointTests {

    @Test
    public final void pointTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toMultiLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POINT.toPolygonBorder();
      });
    }

    @Test
    public final void pointTests() {
      assertArrayEquals(POINT_COORDS, POINT.coordinates, 0);
      assertEquals(POINT_LOC, POINT.toPoint());
      assertTrue(POINT_LOC.equals(POINT.toPoint()));
      assertEquals(Type.POINT, POINT.type);

      Geometry<double[]> actual = Geometry.point(POINT_LOC);
      assertArrayEquals(POINT_COORDS, actual.coordinates, 0);
      assertEquals(POINT_LOC, actual.toPoint());
      assertTrue(POINT_LOC.equals(actual.toPoint()));
      assertEquals(Type.POINT, actual.type);
    }
  }

  @Nested
  @DisplayName("LineString Tests")
  public static class LineStringTests {

    @Test
    public final void lineStringTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toMultiLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        LINE_STRING.toPolygonBorder();
      });
    }

    @Test
    public final void lineStringTests() {
      assertArrayEquals(LINE_COORDS, LINE_STRING.coordinates);
      assertEquals(LINE_LOCS, LINE_STRING.toLineString());
      assertTrue(LINE_LOCS.equals(LINE_STRING.toLineString()));
      assertEquals(Type.LINE_STRING, LINE_STRING.type);

      Geometry<double[][]> actual = Geometry.lineString(LINE_LOCS);
      assertArrayEquals(LINE_COORDS, actual.coordinates);
      assertEquals(LINE_LOCS, actual.toLineString());
      assertTrue(LINE_LOCS.equals(actual.toLineString()));
      assertEquals(Type.LINE_STRING, actual.type);
    }
  }

  @Nested
  @DisplayName("MultiLineString Tests")
  public static class MultiLineStringTests {

    @Test
    public final void multiLineStringTestUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        MULTI_LINE_STRING.toPolygonBorder();
      });
    }

    @Test
    public final void multiLineStringTests() {
      assertArrayEquals(MULTI_LINE_COORDS, MULTI_LINE_STRING.coordinates);
      assertEquals(MULTI_LINE_LOCS, MULTI_LINE_STRING.toMultiLineString());
      assertTrue(MULTI_LINE_LOCS.equals(MULTI_LINE_STRING.toMultiLineString()));
      assertEquals(Type.MULTI_LINE_STRING, MULTI_LINE_STRING.type);

      Geometry<double[][][]> actual = Geometry.multiLineString(MULTI_LINE_LOCS);
      assertArrayEquals(MULTI_LINE_COORDS, actual.coordinates);
      assertEquals(MULTI_LINE_LOCS, actual.toMultiLineString());
      assertTrue(MULTI_LINE_LOCS.equals(actual.toMultiLineString()));
      assertEquals(Type.MULTI_LINE_STRING, actual.type);
    }
  }

  @Nested
  @DisplayName("Polygon Tests")
  public static class PolygonTests {

    @Test
    public final void polygonTestUOE_toPoint() {
      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toPoint();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toLineString();
      });

      assertThrows(UnsupportedOperationException.class, () -> {
        POLYGON.toMultiLineString();
      });
    }

    @Test
    public final void polygonTests() {
      assertEquals(Type.POLYGON, POLYGON.type);

      Geometry<double[][][]> actual = Geometry
          .polygon(MULTI_LINE_LOCS.get(0), MULTI_LINE_LOCS.get(1));
      assertEquals(MULTI_LINE_LOCS.get(0), actual.toPolygonBorder());
      assertTrue(MULTI_LINE_LOCS.get(0).equals(actual.toPolygonBorder()));
      assertEquals(Type.POLYGON, actual.type);
    }
  }

}
