package gov.usgs.earthquake.nshmp.internal;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.internal.NshmFaultSection.PropertyKey;

public class NshmFaultSectionTests {
  private static final String ALT_ID = "455n";
  private static final String ALT_ID_NAME = "qFaultId";
  private static final double DIP = 60.0;
  private static final int ID = 100;
  private static final double LOWER_DEPTH = 15.0;
  private static final String NAME = "Cheraw";
  private static final double RAKE = 45.5;
  private static final double RATE = 0.15;
  private static final RateType RATE_TYPE = RateType.SLIP_RATE;
  private static final String SECONDARY_STATE = "CA";
  private static final String STATE = "CO";
  private static List<LocationList> TRACES = new ArrayList<>();
  private static final double UPPER_DEPTH = 0.0;

  private static String GEOJSON_SUFFIX = ".geojson";
  private static String ZONE = "zone.geojson";

  private static NshmFaultSection FAULT_SECTION;

  static {
    TRACES.add(LocationList.builder()
        .add(38.43087, -103.22463)
        .add(38.40194, -103.23982)
        .build());

    FAULT_SECTION = NshmFaultSection.builder()
        .altId(ALT_ID_NAME, ALT_ID)
        .dip(DIP)
        .id(ID)
        .lowerDepth(LOWER_DEPTH)
        .name(NAME)
        .rake(RAKE)
        .rate(RATE)
        .rateType(RATE_TYPE)
        .secondaryState(SECONDARY_STATE)
        .state(STATE)
        .traces(TRACES)
        .upperDepth(UPPER_DEPTH)
        .toFaultSection();
  }

  @ParameterizedTest(name = "{index} ==> Path: {0}")
  @MethodSource("faultSections")
  public final void fromGeoJsonTests(Path path) throws IOException {
    Feature feature = GeoJson.from(path).toFeature();
    Properties properties = feature.properties();
    NshmFaultSection actual = NshmFaultSection.fromGeoJson(path);
    Feature actualFeature = actual.toFeature();

    checkFeature(feature, actualFeature);

    for (PropertyKey key : PropertyKey.values()) {
      if (properties.containsKey(key.toString())) {
        checkField(key, properties, actual);
      }
    }

    if (properties.map().containsKey(ALT_ID_NAME) && actual.altId().isPresent()) {
      assertEquals(ALT_ID_NAME, actual.altIdName().get());
      assertEquals(properties.get(ALT_ID_NAME), actual.altId().get());
    }
  }

  @Test
  public final void builderFaultSectionExceptionTests() {
    // bad dip
    assertThrows(IllegalArgumentException.class, () -> {
      NshmFaultSection.builder().dip(180.0);
    });

    // Bad lower depth
    assertThrows(IllegalArgumentException.class, () -> {
      NshmFaultSection.builder().lowerDepth(100.0);
    });

    // Bad upper depth
    assertThrows(IllegalArgumentException.class, () -> {
      NshmFaultSection.builder().upperDepth(100.0);
    });

    // No upper depth
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .toFaultSection();
    });

    // No state
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // No name
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // No lower depth
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // No id
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // lowerDepth < upperDepth
    assertThrows(IllegalStateException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(0.0)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(10.0)
          .toFaultSection();
    });

    // No rake
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // No traces
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });

    // No dip
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toFaultSection();
    });
  }

  @Test
  public final void builderInterfaceFaultSectionExceptionTests() {
    // Dip set
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .dip(DIP)
          .id(ID)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .toFaultSection();
    });

    // Lower depth set
    assertThrows(IllegalStateException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .lowerDepth(LOWER_DEPTH)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .toInterfaceFaultSection();
    });

    // No state
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .name(NAME)
          .rake(RAKE)
          .traces(TRACES)
          .toInterfaceFaultSection();
    });

    // No name
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .toInterfaceFaultSection();
    });

    // Upper depth set
    assertThrows(IllegalStateException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .upperDepth(UPPER_DEPTH)
          .toInterfaceFaultSection();
    });

    // No id
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .traces(TRACES)
          .toInterfaceFaultSection();
    });

    // No rake
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .name(NAME)
          .state(STATE)
          .traces(TRACES)
          .toInterfaceFaultSection();
    });

    // No traces
    assertThrows(NullPointerException.class, () -> {
      NshmFaultSection.builder()
          .id(ID)
          .name(NAME)
          .rake(RAKE)
          .state(STATE)
          .toInterfaceFaultSection();
    });
  }

  @Test
  public final void altIdTests() {
    assertTrue(FAULT_SECTION.altId().isPresent());
    assertEquals(ALT_ID, FAULT_SECTION.altId().get());
  }

  @Test
  public final void dipTests() {
    assertTrue(FAULT_SECTION.dip().isPresent());
    assertEquals(DIP, FAULT_SECTION.dip().get(), 0);
  }

  @Test
  public final void idTests() {
    assertEquals(ID, FAULT_SECTION.id());
  }

  @Test
  public final void lowerDepthTests() {
    assertTrue(FAULT_SECTION.lowerDepth().isPresent());
    assertEquals(LOWER_DEPTH, FAULT_SECTION.lowerDepth().get(), 0);
  }

  @Test
  public final void nameTests() {
    assertEquals(NAME, FAULT_SECTION.name());
    assertTrue(NAME.equals(FAULT_SECTION.name()));
  }

  @Test
  public final void rakeTests() {
    assertEquals(RAKE, FAULT_SECTION.rake(), 0);
  }

  @Test
  public final void rateTests() {
    assertTrue(FAULT_SECTION.rate().isPresent());
    assertEquals(RATE, FAULT_SECTION.rate().get(), 0);
  }

  @Test
  public final void rateTypeTests() {
    assertTrue(FAULT_SECTION.rateType().isPresent());
    assertTrue(RATE_TYPE.equals(FAULT_SECTION.rateType().get()));
    assertFalse(RateType.RECURRENCE.equals(FAULT_SECTION.rateType().get()));
    assertEquals(RATE_TYPE, FAULT_SECTION.rateType().get());
  }

  @Test
  public final void secondaryStateTests() {
    assertTrue(FAULT_SECTION.secondaryState().isPresent());
    assertEquals(SECONDARY_STATE, FAULT_SECTION.secondaryState().get());
  }

  @Test
  public final void stateTests() {
    assertEquals(STATE, FAULT_SECTION.state());
    assertTrue(STATE.equals(FAULT_SECTION.state()));
  }

  @Test
  public final void tracesTests() {
    assertEquals(TRACES, FAULT_SECTION.traces());

    for (int i = 0; i < TRACES.size(); i++) {
      LocationList expectedLocs = TRACES.get(i);
      LocationList actualLocs = FAULT_SECTION.traces().get(i);
      for (int j = 0; j < expectedLocs.size(); j++) {
        Location expected = expectedLocs.get(j);
        Location actual = actualLocs.get(j);

        assertEquals(expected, actual);
        assertEquals(expected.longitude, actual.longitude, 0);
        assertEquals(expected.latitude, actual.latitude, 0);
      }
    }
  }

  @Test
  public final void upperDepthTests() {
    assertTrue(FAULT_SECTION.upperDepth().isPresent());
    assertEquals(UPPER_DEPTH, FAULT_SECTION.upperDepth().get(), 0);
  }

  private static void checkFeature(Feature expected, Feature actual) {
    assertEquals(expected.type(), actual.type());

    switch (actual.type()) {
      case LINE_STRING:
        assertEquals(expected.asLineString(), actual.asLineString());
        break;
      case MULTI_LINE_STRING:
        assertEquals(expected.asMultiLineString(), actual.asMultiLineString());
        break;
      default:
        throw new IllegalStateException("GeoJson type [" + actual.type() + "] no supported");
    }
  }

  private static void checkField(PropertyKey key, Properties properties, NshmFaultSection actual) {
    Object expected = properties.get(key.toString());
    switch (key) {
      case DIP:
        if (actual.dip().isPresent()) {
          assertEquals(
              (Double) expected,
              actual.dip().get(),
              0);
        }
        break;
      case LOWER_DEPTH:
        if (actual.lowerDepth().isPresent()) {
          assertEquals((Double) expected, actual.lowerDepth().get(), 0);
        }
        break;
      case MFD_TREE:
        if (actual.mfdTree().isPresent()) {
          assertEquals(expected, actual.mfdTree().get());
        }
        break;
      case NAME:
        assertEquals((String) expected, actual.name());
        break;
      case RAKE:
        assertEquals(
            (Double) expected,
            actual.rake(),
            0);
        break;
      case RATE:
        if (actual.rate().isPresent()) {
          assertEquals(
              (Double) expected,
              actual.rate().get(),
              0);
        }
        break;
      case UPPER_DEPTH:
        if (actual.upperDepth().isPresent()) {
          assertEquals((Double) expected, actual.upperDepth().get(), 0);
        }
        break;
      case RATE_TYPE:
        if (actual.dip().isPresent()) {
          assertEquals(
              RateType.valueOf((String) expected),
              actual.rateType().get());
        }
        break;
      case STATE:
        if (expected.getClass().isInstance(new ArrayList<String>())) {
          String[] states = properties.get(key.toString(), String[].class);
          assertEquals(states[1], actual.secondaryState().get());
        } else {
          assertEquals((String) expected, actual.state());
        }
        break;
      default:
        throw new RuntimeException("Parameter [" + key + "] not allowed");
    }
  }

  private static Stream<Path> faultSections() throws IOException {
    Path path = Paths.get("libs/nshm-fault-sections");
    FaultFinder finder = new FaultFinder();
    Files.walkFileTree(path, finder);
    return finder.paths.stream();
  }

  private static class FaultFinder extends SimpleFileVisitor<Path> {
    List<Path> paths;

    FaultFinder() {
      paths = new ArrayList<>();
    }

    @Override
    public FileVisitResult visitFile(Path path, BasicFileAttributes attrs) {
      Path fileName = path.getFileName();

      if (path.toString().contains(GEOJSON_SUFFIX) && !fileName.toString().contentEquals(ZONE)) {
        paths.add(path);
      }

      return FileVisitResult.CONTINUE;
    }
  }

}
