package gov.usgs.earthquake.nshmp.tree;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.List;

import org.junit.jupiter.api.Test;

import com.google.common.base.CaseFormat;
import com.google.common.collect.ImmutableList;

@SuppressWarnings("javadoc")
public class LogicTreeTests {

  private static final String[] STRING_KEYS = new String[] {
      "Branch0",
      "Branch1",
      "Branch2",
      "Branch3" };

  private static enum EnumKey {
    BRANCH_0,
    BRANCH_1,
    BRANCH_2,
    BRANCH_3;

    @Override
    public String toString() {
      return CaseFormat.UPPER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, name());
    }
  }

  private static final double[] WEIGHTS = new double[] { 0.40, 0.30, 0.20, 0.10 };

  private static final double[] PROBABILITIES =
      new double[] { 0.2, 0.4, 0.41, 0.7, 0.71, 0.9, 0.91, 1.0 };

  @Test
  public final void builderTests() {

    BranchValue bv0 = new BranchValue(0);
    BranchValue bv1 = new BranchValue(1);
    BranchValue bv2 = new BranchValue(2);
    BranchValue bv3 = new BranchValue(3);

    /* RegularTree */

    LogicTree<BranchValue> regularTree = LogicTree.<BranchValue> builder()
        .addBranch(STRING_KEYS[0], bv0, WEIGHTS[0])
        .addBranch(STRING_KEYS[1], bv1, WEIGHTS[1])
        .addBranch(STRING_KEYS[2], bv2, WEIGHTS[2])
        .addBranch(STRING_KEYS[3], bv3, WEIGHTS[3])
        .build();
    assertTrue(regularTree.branches() instanceof ImmutableList);

    Branch<BranchValue> regularBranch = regularTree.branches().get(2);
    assertEquals(STRING_KEYS[2], regularBranch.id());
    assertEquals(bv2, regularBranch.value());
    assertEquals(0.20, regularBranch.weight());
    assertEquals(
        branchString(STRING_KEYS[2], WEIGHTS[2]),
        regularBranch.toString());

    /* EnumTree */

    LogicTree<BranchValue> enumTree = LogicTree.<EnumKey, BranchValue> enumBuilder()
        .addBranch(EnumKey.BRANCH_0, bv0, WEIGHTS[0])
        .addBranch(EnumKey.BRANCH_1, bv1, WEIGHTS[1])
        .addBranch(EnumKey.BRANCH_2, bv2, WEIGHTS[2])
        .addBranch(EnumKey.BRANCH_3, bv3, WEIGHTS[3])
        .build();
    assertTrue(enumTree.branches() instanceof ImmutableList);

    Branch<BranchValue> enumBranch = enumTree.branches().get(2);

    assertEquals(EnumKey.BRANCH_2.name(), enumBranch.id());
    assertEquals(bv2, enumBranch.value());
    assertEquals(0.20, enumBranch.weight());
    assertEquals(
        branchString(EnumKey.BRANCH_2.name(), WEIGHTS[2]),
        enumBranch.toString());

    /* EnumValueTree */

    LogicTree<EnumKey> enumValueTree = LogicTree.<EnumKey> enumValueBuilder()
        .addBranch(EnumKey.BRANCH_0, WEIGHTS[0])
        .addBranch(EnumKey.BRANCH_1, WEIGHTS[1])
        .addBranch(EnumKey.BRANCH_2, WEIGHTS[2])
        .addBranch(EnumKey.BRANCH_3, WEIGHTS[3])
        .build();
    assertTrue(enumValueTree.branches() instanceof ImmutableList);

    Branch<EnumKey> enumValueBranch = enumValueTree.branches().get(2);

    assertEquals(EnumKey.BRANCH_2.name(), enumValueBranch.id());
    assertEquals(EnumKey.BRANCH_2, enumValueBranch.value());
    assertEquals(0.20, enumValueBranch.weight());
    assertEquals(
        branchString(EnumKey.BRANCH_2.name(), WEIGHTS[2]),
        enumValueBranch.toString());

    /* StringValueTree */

    LogicTree<String> stringValueTree = LogicTree.stringValueBuilder()
        .addBranch(STRING_KEYS[0], WEIGHTS[0])
        .addBranch(STRING_KEYS[1], WEIGHTS[1])
        .addBranch(STRING_KEYS[2], WEIGHTS[2])
        .addBranch(STRING_KEYS[3], WEIGHTS[3])
        .build();
    assertTrue(stringValueTree.branches() instanceof ImmutableList);

    Branch<String> stringValueBranch = stringValueTree.branches().get(2);

    assertEquals(STRING_KEYS[2], stringValueBranch.id());
    assertEquals(STRING_KEYS[2], stringValueBranch.value());
    assertEquals(0.20, stringValueBranch.weight());
    assertEquals(
        branchString(STRING_KEYS[2], WEIGHTS[2]),
        stringValueBranch.toString());
  }

  private static String branchString(String id, double weight) {
    return id + '[' + weight + ']';
  }

  @Test
  public final void builderExceptionTests() {

    /* Empty tree: IllegalState, in RegularTree constructor */
    assertThrows(IllegalStateException.class, () -> {
      LogicTree.builder().build();
    });

    /* Weight sum 1 != 1.0, in RegularTree constructor */
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.builder()
          .addBranch(STRING_KEYS[0], new BranchValue(0), WEIGHTS[0])
          .build();
    });

    /* Empty branch id */
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.builder()
          .addBranch(" ", new BranchValue(0), WEIGHTS[0]);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      LogicTree.stringValueBuilder()
          .addBranch(" ", WEIGHTS[0]);
    });

    /* Used builder */
    assertThrows(IllegalStateException.class, () -> {
      LogicTree.Builder<BranchValue> builder = LogicTree.<BranchValue> builder()
          .addBranch(STRING_KEYS[0], new BranchValue(0), 1.0);
      builder.build();
      builder.build();
    });
  }

  @Test
  public final void iteratorTest() {
    LogicTree<BranchValue> tree = LogicTree.<BranchValue> builder()
        .addBranch(STRING_KEYS[0], new BranchValue(1), 1.0)
        .build();
    assertEquals(tree.branches().get(0), tree.iterator().next());
  }

  @Test
  public final void samplingTest() {

    // weights: { 0.40, 0.30, 0.20, 0.10 }
    // probabilities: { 0.2, 0.4, 0.41, 0.7, 0.71, 0.9, 0.91, 1.0 }

    BranchValue bv0 = new BranchValue(0);
    BranchValue bv1 = new BranchValue(1);
    BranchValue bv2 = new BranchValue(2);
    BranchValue bv3 = new BranchValue(3);

    LogicTree<BranchValue> tree = LogicTree.<BranchValue> builder()
        .addBranch(STRING_KEYS[0], bv0, WEIGHTS[0])
        .addBranch(STRING_KEYS[1], bv1, WEIGHTS[1])
        .addBranch(STRING_KEYS[2], bv2, WEIGHTS[2])
        .addBranch(STRING_KEYS[3], bv3, WEIGHTS[3])
        .build();

    List<Branch<BranchValue>> branches = tree.branches();

    List<Branch<BranchValue>> expected = ImmutableList.of(
        branches.get(0),
        branches.get(1),
        branches.get(1),
        branches.get(2),
        branches.get(2),
        branches.get(3),
        branches.get(3),
        branches.get(3));

    List<Branch<BranchValue>> samples = tree.sample(PROBABILITIES);
    assertEquals(expected, samples);

    /*
     * Creating the cumulative weight array introduces double precision math
     * errors such that the test weights herein become [0.4, 0.7,
     * 0.8999999999999999, 0.9999999999999999]. We therefore give package
     * exposure to the cumulativeWeights() method which returns a copy of the
     * cumulative weight array and test that the expected rounding has occurred.
     */
    assertArrayEquals(
        new double[] { 0.4, 0.7, 0.9, 1.0 },
        ((RegularTree<BranchValue>) tree).cumulativeWeights());
  }

  private static class BranchValue {

    @SuppressWarnings("unused")
    int value;

    BranchValue(int value) {
      this.value = value;
    }
  }

}
